local v3 = {}
function v3.new(x, y, z)
	if x == nil then
		return {
			x = 0,
			y = 0,
			z = 0
		}
	end
	if type(x) == "table" then
		return {
			x = x.x,
			y = x.y,
			z = x.z,
		}
	end
	return {
		x = x,
		y = y,
		z = z,
	}
end
function v3.floor(v)
	return {
		x = math.floor(v.x),
		y = math.floor(v.y),
		z = math.floor(v.z),
	}
end
function v3.cmp(v, w)
	return (
		v.x == w.x and
		v.y == w.y and
		v.z == w.z
	)
end
function v3.add(v, w)
	return {
		x = v.x + w.x,
		y = v.y + w.y,
		z = v.z + w.z,
	}
end
function v3.sub(v, w)
	return {
		x = v.x - w.x,
		y = v.y - w.y,
		z = v.z - w.z,
	}
end
function v3.mul(v, a)
	return {
		x = v.x * a,
		y = v.y * a,
		z = v.z * a,
	}
end
function v3.len(v)
	return math.sqrt(
		math.pow(v.x, 2) +
		math.pow(v.y, 2) +
		math.pow(v.z, 2)
	)
end
function v3.norm(v)
	return v3.mul(v, 1.0 / v3.len(v))
end
function v3.distance(v, w)
	return math.sqrt(
		math.pow(v.x - w.x, 2) +
		math.pow(v.y - w.y, 2) +
		math.pow(v.z - w.z, 2)
	)
end
function v3.rotate_y(v, a)
	return {
		x = v.x * math.cos(a) - v.z * math.sin(a),
		y = v.y,
		z = v.x * math.sin(a) + v.z * math.cos(a),
	}
end

local function node_is_solid(n)
	return (n and n.name and minetest.registered_nodes[n.name] and
			minetest.registered_nodes[n.name].walkable)
end

local function isnan(x)
	return x ~= x
end

minetest.register_node("blockmobs:spikeball", {
	drawtype = "nodebox",
	paramtype = "light",
    sunlight_propagates = true,
	tiles = {
		"blockmobs_spikeball.png",
	},
	node_box = {
		type = "fixed",
		fixed = {
			{-0.2, -0.2, -0.2, 0.2, 0.2, 0.2},
			{-0.3, -0.1, -0.1, 0.3, 0.1, 0.1},
			{-0.1, -0.3, -0.1, 0.1, 0.3, 0.1},
			{-0.1, -0.1, -0.3, 0.1, 0.1, 0.3},
			{-0.5, -0.05, -0.05, 0.5, 0.05, 0.05},
			{-0.05, -0.5, -0.05, 0.05, 0.5, 0.05},
			{-0.05, -0.05, -0.5, 0.05, 0.05, 0.5},
		},
	},
	groups={cracky=2},
	sounds = default.node_sound_defaults(),
})

minetest.register_node("blockmobs:spikeball_bright", {
	drawtype = "nodebox",
	paramtype = "light",
    sunlight_propagates = true,
	tiles = {
		"blockmobs_spikeball.png^[brighten",
	},
	node_box = {
		type = "fixed",
		fixed = {
			{-0.2, -0.2, -0.2, 0.2, 0.2, 0.2},
			{-0.3, -0.1, -0.1, 0.3, 0.1, 0.1},
			{-0.1, -0.3, -0.1, 0.1, 0.3, 0.1},
			{-0.1, -0.1, -0.3, 0.1, 0.1, 0.3},
			{-0.5, -0.05, -0.05, 0.5, 0.05, 0.05},
			{-0.05, -0.5, -0.05, 0.05, 0.5, 0.05},
			{-0.05, -0.05, -0.5, 0.05, 0.05, 0.5},
		},
	},
	groups={cracky=2},
	sounds = default.node_sound_defaults(),
})

minetest.register_abm({
	nodenames = {"blockmobs:spikeball"},
	--neighbors = {"default:stone", "default:cobble"},
	interval = 1.0,
	chance = 20,
	action = function(pos, node, active_object_count, active_object_count_wider)
		local objs = minetest.env:get_objects_inside_radius(pos, 4)
		if #objs >= 8 then
			return
		end
		local meta = minetest.env:get_meta(pos)
		local staticdata = meta:get_string("staticdata")
		minetest.env:remove_node(pos)
		local o = minetest.env:add_entity(pos, "blockmobs:moving_spikeball", staticdata)
		o:get_luaentity():load_staticdata(staticdata)
	end,
})

--[[
minetest.register_abm({
	nodenames = {"air"},
	neighbors = {"default:stone", "default:cobble", "default:mossycobble"},
	interval = 1.0,
	chance = 10000,
	action = function(pos, node, active_object_count, active_object_count_wider)
		if active_object_count_wider > 0 then
			--minetest.log("action", "Not spawning spikeball at "..minetest.pos_to_string(pos)
			--		.." because active object count is high")
			return
		end
		local light = minetest.env:get_node_light(pos)
		if not light or light > 2 then
			--minetest.log("action", "Not spawning spikeball at "..minetest.pos_to_string(pos)
			--		.." because light is high ("..dump(light)..")")
			return
		end
		local foundwall = false
		local dirs = {
			{x=-1, y=0, z=0},
			{x=1, y=0, z=0},
			{x=0, y=0, z=-1},
			{x=0, y=0, z=1},
		}
		for _,dp in ipairs(dirs) do
			local p1 = v3.add(pos, dp)
			local n1 = minetest.env:get_node(p1)
			if node_is_solid(n1) then
				foundwall = true
				break
			end
		end
		if not foundwall then
			--minetest.log("action", "Not spawning spikeball at "..minetest.pos_to_string(pos)
			--		.." because didn't find wall")
			return
		end
		minetest.log("action", "Spawning spikeball at "..minetest.pos_to_string(pos))
		minetest.env:add_node(pos, {name="blockmobs:spikeball"})
	end,
})
--]]

minetest.register_entity("blockmobs:moving_spikeball", {
	initial_properties = {
		physical = true,
		collisionbox = {-0.3,-0.3,-0.3, 0.3,0.3,0.3},
		visual = "wielditem",
		textures = {},
		visual_size = {x=0.667, y=0.667},
		hp_max = 6,
	},

	time_from_last_punch = 0,
	punched_timer = -1,
	nextp = nil,
	do_move = true,
	move_timer = 0,
	runtime = 0,
	turn_dir = 1,
	lastdir = nil,
	goodpos = nil,
	no_walls_available_timer = 0,
	homepos = nil,
	time_from_turn = 0,

	get_staticdata = function(self)
		return minetest.serialize({
			turn_dir = self.turn_dir,
			lastdir = self.lastdir,
			goodpos = self.goodpos,
			homepos = self.homepos,
		})
	end,

	load_staticdata = function(self, staticdata)
		local data = minetest.deserialize(staticdata)
		if data then
			if data.turn_dir then
				self.turn_dir = data.turn_dir
			end
			if data.lastdir then
				self.lastdir = v3.new(data.lastdir)
			end
			if data.goodpos then
				self.goodpos = v3.new(data.goodpos)
			end
			if data.homepos then
				self.homepos = v3.new(data.homepos)
			end
		end
		if self.goodpos then
			self.object:setpos(self.goodpos)
		end
	end,

	on_activate = function(self, staticdata)
		self:load_staticdata(staticdata)
		--self.object:set_armor_groups({cracky=3})
		self.object:set_armor_groups({immortal=1})
		prop = {
			is_visible = true,
			textures = {"blockmobs:spikeball"},
			automatic_rotate = math.pi * 1.00 * self.turn_dir,
		}
		self.object:set_properties(prop)
		--self.object:remove()
		if not self.homepos then
			self.homepos = self.object:getpos()
		end
	end,

	start_punch_rotation = function(self)
		prop = {
			textures = {"blockmobs:spikeball_bright"},
			automatic_rotate = math.pi * 3.00 * self.turn_dir,
		}
		self.object:set_properties(prop)
		self.punched_timer = 2
	end,

	on_punch = function(self, puncher, time_from_last_punch, tool_capabilities, dir)
		--self:start_punch_rotation()
		self:do_punch(puncher)
	end,

	on_step = function(self, dtime)
		-- For some reason this happens and screws everything if not handled
		if isnan(self.object:getpos().x) then
			minetest.log("error", "blockmobs:moving_spikeball position is NaN; removing")
			self.object:remove()
			return
		end
		self.time_from_turn = self.time_from_turn + dtime
		self.runtime = self.runtime + dtime
		if self.punched_timer > 0 then
			self.punched_timer = self.punched_timer - dtime
			if self.punched_timer <= 0 then
				self.punched_timer = -1
				prop = {
					textures = {"blockmobs:spikeball"},
					automatic_rotate = math.pi * 1.00 * self.turn_dir,
				}
				self.object:set_properties(prop)
			end
		end
		-- If going at much speed, just accelerate downwards
		if v3.len(self.object:getvelocity()) > 2 then
			self.object:setacceleration(v3.new(0,9,0))
			return
		-- Else don't accelerate anywhere
		else
			self.object:setacceleration(v3.new(0,0,0))
		end
		if self.nextp then
			local speed = 1
			local d_at_speed = dtime * speed
			local p0 = self.object:getpos()
			if v3.distance(self.nextp, p0) < d_at_speed then
				self.object:moveto(self.nextp, false)
				self.object:setvelocity(v3.new(0,0,0))
				self.goodpos = self.nextp
				self.nextp = nil
			else
				local nextpdiff = v3.sub(self.nextp, p0)
				if v3.len(nextpdiff) > 2 then
					self.nextp = nil
					self.goodpos = p0
				else
					local add = v3.mul(v3.norm(nextpdiff), d_at_speed)
					local p1 = v3.add(p0, add)
					--self.object:moveto(p1, true)
					self.object:setvelocity(v3.add(v3.mul(v3.norm(nextpdiff), speed), v3.new(0,0,0)))
				end
			end
		end
		if self.do_move and not self.nextp then
			self.move_timer = self.move_timer - dtime
			if self.move_timer <= 0 then
				--self.move_timer = self.move_timer + 0.5
				--self.move_timer = 1
				self.move_timer = 0
				local p0 = self.object:getpos()
				-- Turn back home if far from it
				if v3.distance(p0, self.homepos) > 20 and self.time_from_turn > 5 then
					self.turn_dir = -self.turn_dir
					prop = {
						textures = {"blockmobs:spikeball"},
						automatic_rotate = math.pi * 1.00 * self.turn_dir,
					}
					self.object:set_properties(prop)
					self.time_from_turn = 0
				end
				local possible_dsts = {}
				local nonsolids = {}
				local spikeballs = {}
				local dirs = {
					{x=-1, y=0, z=0},
					{x=1, y=0, z=0},
					{x=0, y=0, z=-1},
					{x=0, y=0, z=1},
				}
				for _,dp in ipairs(dirs) do
					local p1 = v3.add(p0, dp)
					local n1 = minetest.env:get_node(p1)
					if not node_is_solid(n1) then
						table.insert(nonsolids, p1)
					end
					if n1 and n1.name == "blockmobs:spikeball" then
						table.insert(spikeballs, p1)
					end
					if node_is_solid(n1) then
						local p2d = v3.rotate_y(dp, math.pi/2 * self.turn_dir * (-1))
						local p2 = v3.add(p0, p2d)
						local n2 = minetest.env:get_node(p2)
						if not node_is_solid(n2) then
							table.insert(possible_dsts, p2)
						end
					end
				end
				if #possible_dsts == 0 and self.lastdir then
					local p2d = v3.rotate_y(self.lastdir, math.pi/2 * self.turn_dir)
					local p2 = v3.add(p0, p2d)
					local n2 = minetest.env:get_node(p2)
					local p1d = v3.rotate_y(p2d, math.pi/2 * self.turn_dir)
					local p1 = v3.add(p2, p1d)
					local n1 = minetest.env:get_node(p1)
					if not node_is_solid(n2) and node_is_solid(n1) then
						table.insert(possible_dsts, p2)
					end
				end
				---[[
				if (#nonsolids <= 2 or (#nonsolids <= 3 and #spikeballs == 0)) and self.runtime > 3 and math.random(1,5) == 1 then
					minetest.env:set_node(p0, {name="blockmobs:spikeball"})
					minetest.env:get_meta(p0):set_string("staticdata", self:get_staticdata())
					self.object:remove()
				elseif #possible_dsts > 0 then
				--]]
				--[[
				if #possible_dsts > 0 then
				--]]
					local i = math.random(1, #possible_dsts)
					local nextp = possible_dsts[i]
					nextp = v3.floor(v3.add(nextp, v3.new(0.5,0.5,0.5)))
					self.nextp = nextp
				--[[elseif #nonsolids > 0 then
					local i = math.random(1, #nonsolids)
					local nextp = nonsolids[i]
					nextp = v3.floor(v3.add(nextp, v3.new(0.5,0.5,0.5)))
					self.nextp = nextp]]
				end
				if self.nextp then
					self.lastdir = v3.sub(self.nextp, p0)
					self.no_walls_available_timer = 0
				end
				if #nonsolids == #dirs then
					self.no_walls_available_timer = self.no_walls_available_timer + 1
					if self.no_walls_available_timer > 5 then
						minetest.env:set_node(p0, {name="blockmobs:spikeball"})
						minetest.env:get_meta(p0):set_string("staticdata", self:get_staticdata())
						self.object:remove()
						return
					end
				end
			end
		end
		self.time_from_last_punch = self.time_from_last_punch + dtime
		if self.time_from_last_punch < 0.5 then
			return
		end
		local p0 = self.object:getpos()
		for _,obj in ipairs(minetest.env:get_objects_inside_radius(p0, 3.0)) do
			local obj_entityname = ""
			local obj_entity = obj:get_luaentity()
			if obj_entity then
				obj_entityname = obj_entity.name
			end
			if obj ~= self.object and obj_entityname ~= "blockmobs:moving_spikeball" then
				--print("object inside radius at "..minetest.pos_to_string(obj:getpos()))
				self:do_punch(obj)
			end
		end
	end,
	do_punch = function(self, obj)
		local p0 = self.object:getpos()
		local p1 = obj:getpos()
		if obj:is_player() then
			p1 = v3.add(p1, v3.new(0,1,0))
		end
		local damage = 2
		local toolcap = {
			full_punch_interval=4.0,
			groupcaps={
				fleshy={times={[1]=1/damage*3, [2]=1/damage*2, [3]=1/damage}},
				snappy={times={[1]=1/damage*3, [2]=1/damage*2, [3]=1/damage}},
			}
		}
		local dir = {
			x = p1.x - p0.x,
			y = p1.y - p0.y,
			z = p1.z - p0.z
		}
		dir = v3.norm(dir)
		minetest.log("action", "Punching "..obj:get_player_name())
		obj:punch(self.object, self.time_from_last_punch, toolcap, dir)
		self:start_punch_rotation()
		self.time_from_last_punch = 0
		obj:setvelocity(v3.mul(dir, 8))
		for punchdistance_i=1,5 do
			local punchmove = v3.mul(dir, 3 - 0.5 * punchdistance_i)
			if punchdistance_i % 2 == 1 then
				punchmove.y = punchmove.y + 1
			end
			local p2 = v3.add(p1, punchmove)
			local p3 = v3.add(p2, v3.new(0,0.5,0))
			if not node_is_solid(minetest.env:get_node(p2)) and
					not node_is_solid(minetest.env:get_node(p3)) then
				if obj:is_player() then
					p2 = v3.add(p2, v3.new(0,-1,0))
				end
				obj:setpos(p2)
				break
			end
		end
	end,
})

minetest.register_node("blockmobs:spikecrap", {
	tiles = {
		{image = "blockmobs_spikecrap.png", animation={type="vertical_frames", aspect_w=16, aspect_h=16, length=2.0}},
	},
	groups={cracky=2},
	sounds = default.node_sound_defaults(),
})

minetest.register_abm({
	nodenames = {"blockmobs:spikecrap"},
	interval = 1.0,
	chance = 10,
	action = function(pos, node)
		local objs = minetest.env:get_objects_inside_radius(pos, 5)
		if #objs >= 4 then
			return
		end
		for dx=-1,1 do
		for dz=-1,1 do
			local p1 = v3.add(pos, v3.new(dx,0,dz))
			local n1 = minetest.env:get_node(p1)
			if not node_is_solid(n1) then
				minetest.log("action", "Spawning spikeball at "..minetest.pos_to_string(p1))
				minetest.env:add_node(p1, {name="blockmobs:spikeball"})
				return
			end
		end
		end
	end,
})

minetest.register_on_generated(function(minp, maxp, seed)
	--generate_ore("blockmobs:spikecrap", "default:gravel", minp, maxp, seed+0, 1/12/12/12, 2, 1, -31000,  31000)
	if minp.y > 3 then
		return
	end
	local area = (maxp.x-minp.x+1)*(maxp.z-minp.z+1)
	local pr = PseudoRandom(seed+932)
	spikecrap_level_y = pr:next(minp.y, maxp.y)
	local is_empty = false
	local light = nil
	local last_empty = false
	local last_light = nil
	local possible_spikecraps = {}
	local d = 4
	for x=minp.x/d+1,maxp.x/d-1 do
		local last_node_known = false
	for z=minp.z+1,maxp.z-1 do
		local p = v3.new(x*d, spikecrap_level_y, z)
		local n = minetest.env:get_node(p)
		last_empty = is_empty
		last_light = light
		is_empty = not node_is_solid(n)
		light = minetest.env:get_node_light(p, 0.5)
		if last_node_known then
			local useful_light = nil
			if is_empty then
				useful_light = light
			else
				useful_light = last_light
			end
			if is_empty and not last_empty and useful_light <= 2 then
				local p1 = v3.add(p, v3.new(0, 0, -1))
				local p2 = v3.add(p1, v3.new(0, 2, -1))
				local p3 = v3.add(p1, v3.new(0, 2, -8))
				local p4 = v3.add(p1, v3.new(-3, 1, -2))
				local p5 = v3.add(p1, v3.new(3, 1, -2))
				if node_is_solid(minetest.env:get_node(p2)) and
					node_is_solid(minetest.env:get_node(p3)) and
					node_is_solid(minetest.env:get_node(p4)) and
					node_is_solid(minetest.env:get_node(p5))
				then
					local spikecrap = {
						p = p1,
						dir = {x=0, y=0, z=-1},
					}
					table.insert(possible_spikecraps, spikecrap)
				end
			end
			if last_empty and not is_empty and useful_light <= 2 then
				local p1 = v3.add(p, v3.new(0, 0, 0))
				local p2 = v3.add(p1, v3.new(0, 2, 1))
				local p3 = v3.add(p1, v3.new(0, 2, 8))
				local p4 = v3.add(p1, v3.new(-3, 1, 2))
				local p5 = v3.add(p1, v3.new(3, 1, 2))
				if node_is_solid(minetest.env:get_node(p2)) and
					node_is_solid(minetest.env:get_node(p3)) and
					node_is_solid(minetest.env:get_node(p4)) and
					node_is_solid(minetest.env:get_node(p5))
				then
					local spikecrap = {
						p = p1,
						dir = {x=0, y=0, z=1},
					}
					table.insert(possible_spikecraps, spikecrap)
				end
			end
		end
		last_node_known = true
	end
	end
	local num_spikecraps_to_generate = area/200
	if num_spikecraps_to_generate > #possible_spikecraps then
		num_spikecraps_to_generate = #possible_spikecraps
	end
	local pr = PseudoRandom(seed+9452)
	for spikecraps_generate_i=1,num_spikecraps_to_generate do
		local spikecrap_i = pr:next(1, #possible_spikecraps)
		local spikecrap = possible_spikecraps[spikecrap_i]
		minetest.log("action", "Spikecrap at "..minetest.pos_to_string(spikecrap.p))
		minetest.env:set_node(spikecrap.p, {name="blockmobs:spikecrap"})
	end
end)
